//! Custom error type representing all possible error variants for peach-web.

use peach_lib::error::PeachError;
use peach_lib::{serde_json, serde_yaml};
use serde_json::error::Error as JsonError;
use serde_yaml::Error as YamlError;

/// Custom error type encapsulating all possible errors for the web application.
#[derive(Debug)]
pub enum PeachWebError {
    Json(JsonError),
    Yaml(YamlError),
    FailedToRegisterDynDomain(String),
    PeachLib { source: PeachError, msg: String },
}

impl std::error::Error for PeachWebError {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match *self {
            PeachWebError::Json(ref source) => Some(source),
            PeachWebError::Yaml(ref source) => Some(source),
            PeachWebError::FailedToRegisterDynDomain(_) => None,
            PeachWebError::PeachLib { ref source, .. } => Some(source),
        }
    }
}

impl std::fmt::Display for PeachWebError {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        match *self {
            PeachWebError::Json(ref source) => write!(f, "Serde JSON error: {}", source),
            PeachWebError::Yaml(ref source) => write!(f, "Serde YAML error: {}", source),
            PeachWebError::FailedToRegisterDynDomain(ref msg) => {
                write!(f, "DYN DNS error: {}", msg)
            }
            PeachWebError::PeachLib { ref source, .. } => write!(f, "{}", source),
        }
    }
}

impl From<JsonError> for PeachWebError {
    fn from(err: JsonError) -> PeachWebError {
        PeachWebError::Json(err)
    }
}

impl From<YamlError> for PeachWebError {
    fn from(err: YamlError) -> PeachWebError {
        PeachWebError::Yaml(err)
    }
}

impl From<PeachError> for PeachWebError {
    fn from(err: PeachError) -> PeachWebError {
        PeachWebError::PeachLib {
            source: err,
            msg: "".to_string(),
        }
    }
}
